# -*- coding: iso-8859-1 -*-
# --------------------------------------------------------------------
# LST2005 -- Python Modul zur Berechnung der Deutschen Lohnsteuer 2005
# $Id: LST2005.py 37 2005-07-20 14:18:07Z wilde $
# --------------------------------------------------------------------
#
# Copyright (c) 2005 by Intevation GmbH
# Authors:
# Sascha Wilde <wilde@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with this package for details.

"""Lohnsteuerberechnung nach dem offiziellen Programmablaufplan
   wie im offiziellen Programmablaufplan des Bundesfinanzministerium
   dokumentiert.
   http://www.bundesfinanzministerium.de/Service/Interaktiver-Steuerrechner-.577.htm"""

__version__ = "$Revision: 37 $"
# $Source$

def _ModulVersion():
    return __version__[11:-2]

# Die Variablen Namen sind hsslich, die Algorithmen nicht
# dokumentiert, der Code grausam -- dafr entspricht alles Zeile fr
# Zeile obigem Dokument.  Jedenfalls sollte es so sein...

from math import ceil, floor

# Wir brauchen eigene Rundungsfunktionen mit definierter Przision
def FixedPointFloor(value, precision=2):
    i = 10.0 ** precision
    return floor(value * i) / i

def FixedPointCeil(value, precision=2):
    i = 10.0 ** precision
    return ceil(value * i) / i

# Ein paar Konstanten um das ganze etwas Benutzbarer zu machen:
JAHR  = 1
MONAT = 2
WOCHE = 3
TAG   = 4
I   = 1
II  = 2
III = 3
IV  = 4
V   = 5
VI  = 6

class LST:
    def __init__(self,
                 RE4=0,
                 ALTER1=0,
                 HINZUR=0,
                 JFREIB=0,
                 JHINZU=0,
                 JRE4  =0,
                 JVBEZ =0,
                 KRV   =0,
                 LZZ   =2,
                 R     =0,
                 SONSTB=0,
                 STKL  =1,
                 VBEZ  =0,
                 VBEZM =0,
                 VBEZS =0,
                 VBS   =0,
                 VMT   =0,
                 WFUNDF=0,
                 ZKF   =0,
                 ZMVB  =0):
        self.Set_RE4(RE4)
        self.Set_ALTER1(ALTER1)
        self.Set_HINZUR(HINZUR)
        self.Set_JFREIB(JFREIB)
        self.Set_JHINZU(JHINZU)
        self.Set_JRE4(JRE4)
        self.Set_JVBEZ(JVBEZ)
        self.Set_KRV(KRV)
        self.Set_LZZ(LZZ)
        self.Set_R(R)
        self.Set_SONSTB(SONSTB)
        self.Set_STKL(STKL)
        self.Set_VBEZ(VBEZ)
        self.Set_VBEZM(VBEZM)
        self.Set_VBEZS(VBEZS)
        self.Set_VBS(VBS)
        self.Set_VMT(VMT)
        self.Set_WFUNDF(WFUNDF)
        self.Set_ZKF(ZKF)
        self.Set_ZMVB(ZMVB)
        # Vorgegebene Ausgangsparameter
        self.BK = 0
        self.BKS = 0
        self.BKV = 0
        self.LSTLZZ = 0
        self.SOLZLZZ = 0
        self.SOLZS = 0
        self.SOLZV = 0
        self.STS = 0
        self.STV = 0

    def Calc(self):
        """Berechnet die Lohnsteuer,
        setzt BK, BKS, BKV, LSTLZZ, SOLZLZZ, SOLZS, SOLZV, STS und STV"""
        # Interne Felder
        # Laut Dokumentation sollen diese vor der Berechnung gelscht werden,
        # es ist mir nicht klar, ob das wirklich ntig ist...
        self._ALTE = 0
        self._ANP = 0
        self._ANTEIL1 = 0
        self._ANTEIL2 = 0
        self._BMG = 0
        self._DIFF = 0
        self._EFA = 0
        self._FVB = 0
        self._FVBZ = 0
        self._JBMG = 0
        self._JW = 0
        self._KENNZ = 0
        self._KFB = 0
        self._KZTAB = 1
        self._LSTJAHR = 0
        self._LST1 = 0
        self._LST2 = 0
        self._LST3 = 0
        self._MIST = 0
        self._RE4LZZ = 0
        self._RE4LZZV = 0
        self._RW = 0.0
        self._SAP = 0
        self._SOLZFREI = 0
        self._SOLZJ = 0.0  # 2 Dezimalstellen
        self._SOLZMIN = 0.0  # 2 Dezimalstellen
        self._ST = 0
        self._ST1 = 0
        self._ST2 = 0
        self._VBEZB = 0
        self._VHB = 0
        self._VSP = 0.0  # 2 Dezimalstellen
        self._VSPN = 0
        self._VSP1 = 0.0  # 2 Dezimalstellen
        self._VSP2 = 0.0  # 2 Dezimalstellen
        self._VSPKURZ = 0
        self._VSPMAX1 = 0
        self._VSPMAX2 = 0
        self._VSPO = 0.0  # 2 Dezimalstellen
        self._VSPREST = 0.0  # 2 Dezimalstellen
        self._VSPVOR = 0.0  # 2 Dezimalstellen
        self._X = 0.0  # 2 Dezimalstellen
        self._Y = 0.0  # 6 Dezimalstellen
        self._ZRE4 = 0.0  # 2 Dezimalstellen
        self._ZRE4VP = 0.0  # 2 Dezimalstellen
        self._ZRE4VP1 = 0.0  # 2 Dezimalstellen
        self._ZTABFB = 0
        self._ZVE = 0
        self._ZX = 0
        self._ZZX = 0
        self._HOCH = 0
        self._VERGL = 0
        # ------------------------------------------------------------
        # Anfang der Berechnung
        self._MRE4LZZ()
        self._RE4LZZ = self.RE4 - self._FVB \
                        - self._ALTE - self.WFUNDF \
                        + self.HINZUR
        self._RE4LZZV = self.RE4 - self._FVB - self._ALTE
        self._MRE4()
        self._MZTABFB()
        self._MLSTJAHR()
        self._LSTJAHR = self._ST
        self._JW = self._LSTJAHR * 100
        self._UPANTEIL()
        self.LSTLZZ = self._ANTEIL1
        if self.ZKF > 0:
            self._ZTABFB += self._KFB
            self._MLSTJAHR()
            self._JBMG = self._ST
        else:
            self._JBMG = self._LSTJAHR
        self._MSOLZ()
        self._MSONST()
        self._MVMT()
        
    # Benutzte Unterprogramme:
    def _MRE4LZZ(self):
        if self.VBEZ == 0:
            self._FVBZ = 0
            self._FVB = 0
        else:
            if self.LZZ == 1:
                self._VBEZB = self.VBEZM * self.ZMVB + self.VBEZS
                self._FVBZ = 75 * self.ZMVB
            else:
                self._VBEZB = self.VBEZM * 12 + self.VBEZS
                self._FVBZ = 900
            self._FVB = ceil(self._VBEZB * 0.4)
            if self._FVB > 300000:
                self._FVB = 300000
            self._JW = self._FVB
            self._UPANTEIL()
            self._FVB = self._ANTEIL2
        if self.ALTER1 == 0:
            self._ALTE = 0
        else:
            self._BMG = self.RE4 - self.VBEZ
            self._ALTE = ceil(self._BMG * 0.4)
            self._JW = 190000
            self._UPANTEIL()
            if self._ALTE > self._ANTEIL2:
                self._ALTE = self._ANTEIL2

    def _MRE4(self):
        if self.LZZ == 1:
            self._ZRE4 = FixedPointFloor(self._RE4LZZ / 100.0)
            self._ZRE4VP = FixedPointFloor(self._RE4LZZV / 100.0)
        elif self.LZZ == 2:
            self._ZRE4 = FixedPointFloor((self._RE4LZZ + 0.67) * 0.12)
            self._ZRE4VP = FixedPointFloor((self._RE4LZZV + 0.67) * 0.12)
        elif self.LZZ == 3:
            self._ZRE4 = FixedPointFloor((self._RE4LZZ + 0.89) * 3.6 / 7)
            self._ZRE4VP = FixedPointFloor((self._RE4LZZV + 0.89) * 3.6 / 7)
        else:
            self._ZRE4 = FixedPointFloor((self._RE4LZZ + 0.56) * 3.6)
            self._ZRE4VP = FixedPointFloor((self._RE4LZZV + 0.56) * 3.6)
        if self._ZRE4 < 0:
            self._ZRE4 = 0.0

    def _MZTABFB(self):
        self._KZTAB = 1
        if self.STKL < 6:
            if self.VBEZ > 0:
                self._ANP = 102
            if self.RE4 > self.VBEZ:
                self._ANP += 920
        else:
            self._ANP = 0            
        if self.STKL == 1:
            self._SAP = 36
            self._KFB = self.ZKF * 5808
        elif self.STKL == 2:
            self._EFA = 1308
            self._SAP = 36
            self._KFB = self.ZKF * 5808
        elif self.STKL == 3:
            self._KZTAB = 2
            self._SAP = 72
            self._KFB = self.ZKF * 5808
        elif self.STKL == 4:
            self._SAP = 36
            self._KFB = self.ZKF * 2904
        else:
            self._KFB = 0
        self._ZTABFB = self._EFA + self._ANP + self._SAP + self._FVBZ
        self._KENNZ = 0

    def _MLSTJAHR(self):
        if self.STKL < 5:
            self._UPEVP()
        else:
            self._VSP = 0.0  
        self._ZVE = self._ZRE4 - self._ZTABFB - self._VSP
        if self._ZVE < 1:
            self._ZVE = self._X = 0.0
        else:
            self._X = floor(float(self._ZVE) / self._KZTAB)
        if self.STKL < 5:
            self._UPTAB05()
        else:
            self._MST5_6()

    def _UPEVP(self):
        if self.KRV == 1:
            self._VSP1 = 1.0
        else:
            if self._ZRE4VP > 62400:
                self._ZRE4VP = 62400
            self._VSP1 = 0.2 * self._ZRE4VP
            self._VSP1 = FixedPointFloor(self._VSP1 * 0.0975)
        self._VSP2 = FixedPointFloor(0.11 * self._ZRE4VP)
        self._VHB = 1500 * self._KZTAB
        if self._VSP2 > self._VHB:
            self._VSP2 = self._VHB
        self._VSPN = ceil(self._VSP1 + self._VSP2)
        self._MVSP()
        if self._VSPN > self._VSP:
            self._VSP = self._VSPN

    def _MVSP(self):
        self._VSPO = self._ZRE4VP * 0.2
        self._VSPVOR = 3068 * self._KZTAB
        self._VSPMAX1 = 1334 * self._KZTAB
        self._VSPMAX2 = 667 * self._KZTAB
        self._VSPKURZ = 1134 * self._KZTAB
        if self.KRV == 1:
            if self._VSPO > self._VSPKURZ:
                self._VSP = self._VSPKURZ
            else:
                self._VSP = ceil(self._VSPO)
        else:
            self._UMVSP()

    def _UMVSP(self):
        if self._KENNZ == 1:
            self._VSPVOR = self._VSPVOR - self._ZRE4VP1 * 0.16
        else:
            self._VSPVOR = self._VSPVOR - self._ZRE4VP * 0.16
        if self._VSPVOR < 0:
            self._VSPVOR = 0.0
        if self._VSPO > self._VSPVOR:
            self._VSP = self._VSPVOR
            self._VSPREST = self._VSPO - self._VSPVOR
            if self._VSPREST > self._VSPMAX1:
                self._VSP += self._VSPMAX1
                self._VSPREST = FixedPointCeil((self._VSPREST - self._VSPMAX1) / 2.0)
                if self._VSPREST > self._VSPMAX2:
                    self._VSP = ceil(self._VSP + self._VSPMAX2)
                else:
                    self._VSP = ceil(self._VSP + self._VSPREST)
            else:
                self._VSP = ceil(self._VSP + self._VSPREST)
        else:
            self._VSP = ceil(self._VSPO)

    def _MST5_6(self):
        self._ZZX = self._X
        if self._ZZX > 25812:
            self._ZX = 25812
            self._UP5_6()
            self._ST = floor(self._ST + (self._ZZX - 25812) * 0.42)
        else:
            self._ZX = self._ZZX
            self._UP5_6()
            if self._ZZX > 9144:
                self._VERGL = self._ST
                self._ZX = 9144
                self._UP5_6()
                self._HOCH = floor(self._ST + (self._ZZX - 9144) * 0.42)
                if self._HOCH < self._VERGL:
                    self._ST = self._HOCH
                else:
                    self._ST = self._VERGL

    def _UP5_6(self):
        self._X = self._ZX * 1.25
        self._UPTAB05()
        self._ST1 = self._ST
        self._X = self._ZX * 0.75
        self._UPTAB05()
        self._ST2 = self._ST
        self._DIFF = (self._ST1 - self._ST2) * 2
        self._MIST = floor(self._ZX * 0.15)
        if self._MIST > self._DIFF:
            self._ST = self._MIST
        else:
            self._ST = self._DIFF

    def _MSOLZ(self):
        self._SOLZFREI = 972 * self._KZTAB
        if self._JBMG > self._SOLZFREI:
            self._SOLZJ = FixedPointFloor(self._JBMG * 5.5 / 100.0)
            self._SOLZMIN = (self._JBMG - self._SOLZFREI) * 20 / 100.0
            if self._SOLZMIN < self._SOLZJ:
                self._SOLZJ = self._SOLZMIN
            self._JW = self._SOLZJ * 100
            self._UPANTEIL()
            self.SOLZLZZ = self._ANTEIL1
        else:
            self.SOLZLZZ = 0
        if self.R > 0:
            self._JW = self._JBMG * 100
            self._UPANTEIL()
            self.BK = self._ANTEIL1
        else:
            self.BK = 0

    def _UPANTEIL(self):
        if self.LZZ == 1:
            self._ANTEIL1 = self._JW
            self._ANTEIL2 = self._JW
        elif self.LZZ == 2:
            self._ANTEIL1 = floor(self._JW / 12.0)
            self._ANTEIL2 = ceil(self._JW / 12.0)
        elif self.LZZ == 3:
            self._ANTEIL1 = floor(self._JW * 7 / 360.0)
            self._ANTEIL2 = ceil(self._JW * 7 / 360.0)
        else:
            self._ANTEIL1 = floor(self._JW / 360.0)
            self._ANTEIL2 = ceil(self._JW / 360.0)            

    def _UPTAB05(self):
        if self._X < 7665:
            self._ST = 0
        elif self._X < 12740:
            self._Y = (self._X - 7664) / 10000.0
            self._RW = self._Y * 883.74
            self._RW += 1500
            self._ST = floor(self._RW * self._Y)
        elif self._X < 52152:
            self._Y = (self._X - 12739) / 10000.0
            self._RW = self._Y * 228.74
            self._RW += 2397
            self._RW *= self._Y
            self._ST = floor(self._RW + 989)
        else:
            self._ST = floor(self._X * 0.42 - 7914)
        self._ST *= self._KZTAB

    def _MSONST(self):
        if self.SONSTB > 0:
            # ------------------------------
            # Nicht im offiziellen Programm-
            # ablaufplan: Attribute sichern
            old_lzz = self.LZZ
            old_vbez = self.VBEZ
            old_re4 = self.RE4
            # ------------------------------
            self.LZZ = 1
            self.VBEZ = self.JVBEZ
            self.RE4 = self.JRE4
            self._MRE4LZZ()
            self._MRE4LZZ2()
            self._MRE4()
            self._MZTABFB()
            self._MLSTJAHR()
            self._LST1 = self._ST * 100
            self.VBEZ = self.JVBEZ + self.VBS
            self.RE4 = self.JRE4 + self.SONSTB
            self._MRE4LZZ()
            self._MRE4LZZ2()
            self._MRE4()
            self._MLSTJAHR()
            self._LST2 = self._ST * 100
            self.STS = self._LST2 - self._LST1
            self.SOLZS = self.STS * 5.5 / 100
            if self.R > 0:
                self.BKS = self.STS
            else:
                self.BKS = 0
            # ------------------------------
            # Nicht im offiziellen Programm-
            # ablaufplan: Attribute
            # wiederherstellen
            self.LZZ = old_lzz
            self.VBEZ = old_vbez
            self.RE4 = old_re4
            # ------------------------------
        else:
            self.STS = 0
            self.SOLZS = 0
            self.BKS = 0

    def _MRE4LZZ2(self):
        self._RE4LZZ = self.RE4 - self._FVB - self._ALTE \
                       - self.JFREIB + self.JHINZU
        self._RE4LZZV = self.RE4 - self._FVB - self._ALTE

    def _MVMT(self):
        if self.VMT > 0:
            # ------------------------------
            # Nicht im offiziellen Programm-
            # ablaufplan: Attribute sichern
            old_lzz = self.LZZ
            old_vbez = self.VBEZ
            old_re4 = self.RE4
            # ------------------------------   
            self.LZZ = 1
            self.VBEZ = self.JVBEZ + self.VBS
            self.RE4 = self.JRE4 + self.SONSTB
            self._MRE4LZZ()
            self._MRE4LZZ2()
            self._MRE4()
            self._MZTABFB()
            self._MLSTJAHR()
            self._LST1 = self._ST * 100
            self.RE4 = self.JRE4 + self.SONSTB + self.VMT
            self._MRE4LZZ()
            self._MRE4LZZ2()
            self._MRE4()
            self._KENNZ = 1
            self._ZRE4VP1 = self._ZRE4VP
            self._MLSTJAHR()
            self._LST3 = self._ST * 100
            self.RE4 = self.JRE4 + self.SONSTB
            self._MRE4LZZ()
            self.RE4 = self.JRE4 + self.SONSTB + self.VMT / 5
            self._MRE4LZZ2()
            self._MRE4()
            self._MLSTJAHR()
            self._LST2 = self._ST * 100
            self.STV = (self._LST2 - self._LST1) * 5
            self._LST3 -= self._LST1
            if self._LST3 < self.STV:
                self.STV = self._LST3
            self.SOLZV = floor(self.STV * 5.5 / 100)
            if self.R > 0:
                self.BKV = self.STV
            else:
                self.BKV = 0
            # ------------------------------
            # Nicht im offiziellen Programm-
            # ablaufplan: Attribute
            # wiederherstellen
            self.LZZ = old_lzz
            self.VBEZ = old_vbez
            self.RE4 = old_re4
            # ------------------------------
        else:
            self.STV = 0
            self.SOLZV = 0
            self.BKV = 0

    # Methoden zum geprften setzen der Wert
    # FIX ME: Prfung _sehr_ unvollstndig
    def Set_RE4(self, value):
        assert value >= 0, "must not be negative"
        self.RE4 = value

    def Set_ALTER1(self, value):
        assert value in (0,1), "must be 0 or 1"
        self.ALTER1 = value
        
    def Set_HINZUR(self, value):
        self.HINZUR = value
        
    def Set_JFREIB(self, value):
        self.JFREIB = value
        
    def Set_JHINZU(self, value):
        self.JHINZU = value
        
    def Set_JRE4(self, value):
        self.JRE4 = value
        
    def Set_JVBEZ(self, value):
        self.JVBEZ = value
        
    def Set_KRV(self, value):
        assert value in (0,1), "must be 0 or 1"
        self.KRV = value
        
    def Set_LZZ(self, value):
        assert value in (1,2,3,4), \
               "must be in range 1-4 (JAHR, MONAT, WOCHE, TAG)"
        self.LZZ = value
        
    def Set_R(self, value):
        assert value >= 0.0 and value <= 100.0, \
               "must be in range 0.0-100.0 (Percent)"
        self.R = value
        
    def Set_SONSTB(self, value):
        self.SONSTB = value
        
    def Set_STKL(self, value):
        assert value in (1,2,3,4,5,6), \
               "must be in range 1-6 (I II III IV V VI)"
        self.STKL = value
        
    def Set_VBEZ(self, value):
        self.VBEZ = value

    def Set_VBEZM(self, value):
        self.VBEZM = value

    def Set_VBEZS(self, value):
        self.VBEZS = value

    def Set_VBS(self, value):
        self.VBS = value

    def Set_VMT(self, value):
        self.VMT = value

    def Set_WFUNDF(self, value):
        self.WFUNDF = value

    def Set_ZKF(self, value):
        assert float(value) == float("%.1f" % value) and \
               value >= 0, \
               "must be positive, and must not have more than one decimal digit"
        self.ZKF = value
        
    def Set_ZMVB(self, value):
        self.ZMVB = value

# --------------------------------------------------------------------
# Eine etwas schnere API
#
# FIX ME: Diese API bercksichtigt nicht alle Mglichen Parameter und
# Berechnungen, es fehlen insbesondere die Berechnungen zu Mehrjhrigen
# Bezgen und Sonstigen Leistungen.

class LStRechner2005(LST):
    def __init__(self):
        LST.__init__(self)
    
    def SetLohn(self, lohn):
        """Setzt Lohn in Euro.Cent"""
        self.Set_RE4(round (lohn * 100.0))

    def GetLohn(self):
        """Liefert Lohn in Euro.Cent"""
        return self.RE4 / 100.0

    def SetZeitraum(self, lzz):
        """Setzt Berechnungszeitraum (JAHR, MONAT, WOCHE, TAG)"""
        self.Set_LZZ(lzz)

    def GetZeitraum(self):
        """Liefert Berechnungszeitraum (JAHR, MONAT, WOCHE, TAG)"""
        return self.LZZ

    def SetSteuerklasse(self, stkl):
        """Setzt Steuerklasse (I, II, III, IV, V, VI)"""
        self.Set_STKL(stkl)

    def GetSteuerklasse(self):
        """Liefert Steuerklasse (I, II, III, IV, V, VI)"""
        return self.STKL

    def SetKirchensteuerProzent(self, prozent):
        """Setzt Kirchensteuer in Prozent,
        0 wenn keine Kirchensteuer zu zahlen ist."""
        self.Set_R(prozent)

    def GetKirchensteuerProzent(self):
        """Liefert Kirchensteuer in Prozent
        oder 0 wenn keine Kirchensteuer zu zahlen ist."""
        return self.R

    def SetKinderfreibetrag(self, kfb):
        """Setzt Kinderfreibetrag lt. Lohnsteuerkarte"""
        self.Set_ZKF(kfb)

    def GetKinderfreibetrag(self):
        """Liefert Kinderfreibetrag lt. Lohnsteuerkarte"""
        return self.ZFK

    def GetLohnsteuer(self):
        """Liefert Lohnsteuer in Euro.Cent"""
        return round(self.LSTLZZ / 100, 2)

    def GetSoli(self):
        """Liefert Solidarittszuschlag in Euro.Cent"""
        return FixedPointFloor(self.SOLZLZZ / 100, 2)

    def GetKirchensteuer(self):
        """Liefert Kirchensteuer in Euro.Cent"""
        return FixedPointFloor(self.BK * self.R / 10000, 2)
